#!/usr/bin/env python3
""" 
  Course:      Data Structures and Algorithms for CL III - WS2324
  Assignment:  (Assignment number)
  Author:      (Your name)
  Description: (...)
  Honor Code:  I pledge that this program represents my own work.

"""
import numpy as np
from collections import Counter

class UnigramModel:
    """This class implements a "unigram language model"

    We assign probabilities to words based on their frequency in the
    training corpus. If the word is unknown, the probability estimate
    falls back to probabilities of letters.

    Understanding this class/model is not crucial for the solution of
    the assignment. You can see the intended use case in the main
    section of the file below. However, you are encouraged to study
    and understand the model, and ask questions if you do not
    understand.
    """
    def __init__(self, training_file=None):
        self.words = dict()
        self.letters = dict()
        if training_file is not None:
            words = []
            with open(training_file, 'rt') as f:
                for line in f:
                    words.extend(line.strip().split())
            self.train(words)

    def train(self, words):
        """Estimate the probabilities from given word list."""
        wordc = Counter(words)
        wtypes, wtokens = len(wordc), sum(wordc.values())
        prob_unk = wtypes/wtokens #first occurrence of each word is unknown
        prob_known = 1 - prob_unk
        for word, freq in wordc.items():
            self.words[word] = np.log2(prob_known) \
                    + np.log2(freq) - np.log2(wtokens)
        self.words['<unk>'] = np.log2(prob_unk)

        charc = Counter()
        for w in self.words:
            charc.update(w)
        ctypes, ctokens  = len(charc), sum(charc.values()) 
        prob_unk = ctypes / ctokens
        prob_known = 1 - prob_unk
        for ch, freq in charc.items():
            self.letters[ch] = np.log2(prob_known) \
                    + np.log2(freq) - np.log2(ctokens)
        self.letters['<unk>'] = np.log2(prob_unk)

    def get_logprob(self, word):
        """Return the log probability of given sequence. """
        if word in self.words:
            return self.words[word]
        else:
            logp = self.words['<unk>']
            for ch in word:
                logp += self.letters.get(ch, self.letters.get('<unk>'))
            return logp


def initialize_graph(s, prob_func):
    """Return the adjacency matrix for a given string.

    This function initializes and returns a 2D numpy array (a matrix),
    which represents the weighted directed acyclic graph for
    segmentation. See the assignment description for details.

    Parameters:
    --------
    s           The input sequence for which the graph will be built.
    prob_func   A function that returns log probabilities for a given
                string. Note that, since we want to minimize the sum
                of the path, we negate the log probability when
                assigning to the appropriate matrix cell.
    """
    pass 

def find_shortest_path(G):
    """Find the shortest path on G using the Bellman-Ford algorithm.

    Parameters:
    ----------
    G   The graph represented as a adjecency matrix.

    Return value: The ordered list of nodes in the shortest path
                  from node 0 to the final node.
    """
    pass

def segment_sequence(s, prob_func):
    """Segment the given sequence s using the prob_func

    This makes use of the above graph initialization and shortest path 
    implementations to segment a string.

    Parameters:
    ----------
    s           The input string to be segmented.
    prob_func   Probability function to pass to initialize_graph().

    Return value: a segmentation of s (another string)
                  where boundaries are indicated with
                  with a single space character.
    """
    pass

# Example usage
if __name__ == '__main__':
    lm = UnigramModel('childes-train.txt')
    G = initialize_graph('theman', prob_func=lm.get_logprob)
    print(G)

    path = find_shortest_path(G)
    print(path)
    
    print(segment_sequence('acat', lm.get_logprob))

# Segment the whole corpus and check its accuracy (this may take some time)
#
# This is not a fair evaluation. Normally we need to test the system
# on "unseen data" On the other hand, exact match is also too strict.
# We often use metrics that give partial credit to the segmentation
# methods (e.g. for finding some boundaries or words).

    from tqdm import tqdm
    with open('childes-train.txt', 'rt') as f:
        sentences = f.read().strip('\n')
    ncorrect = 0
    for i, line in enumerate(tqdm(sentences)):
        line = line.strip()
        if line == segment_sequence(
                     line.replace(' ', ''),lm.get_logprob):
            ncorrect += 1
    print(f"Exact match = {ncorrect/i} ({ncorrect}/{i})")
